library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.axi3intercon_pkg.all;
use work.axi_aw_router_pkg.all;

entity axi3intercon_aw_router is
	port(
		aclk           : in  std_logic;
		rst            : in  std_logic;
		masters_out    : in  axi_aw_masters_out_t(0 to MASTER_COUNT - 1);
		masters_in     : out axi_aw_masters_in_t(0 to MASTER_COUNT - 1);
		slaves_out     : in  axi_aw_slaves_out_t(0 to SLAVE_COUNT);
		slaves_in      : out axi_aw_slaves_in_t(0 to SLAVE_COUNT);
		write_locks    : out write_locks_t(0 to MASTER_COUNT - 1); -- write_* signals come/go from/to the write router.
		write_releases : in  write_release_t;
		address_array  : in  axi_slave_addresses_t(0 to SLAVE_COUNT - 1);
		mask_array     : in  axi_slave_addresses_t(0 to SLAVE_COUNT - 1)
	);
end entity axi3intercon_aw_router;

architecture RTL of axi3intercon_aw_router is
	alias clk is aclk;
	type aw_state_t is (AW_READY, AW_ACTIVE, AW_BLOCK);
	type aw_states_t is array (0 to MASTER_COUNT - 1) of aw_state_t;
	signal aw_states     : aw_states_t;
	signal write_locks_s : write_locks_t(0 to MASTER_COUNT - 1);
	signal awready_s     : std_logic_vector(0 to MASTER_COUNT - 1);
begin
	write_locks <= write_locks_s;

	awready_gen : for i in 0 to MASTER_COUNT - 1 generate
		masters_in(i).awready <= awready_s(i);
	end generate awready_gen;

	aw_router : process(clk, rst) is
		variable slave_idx    : integer range 0 to SLAVE_COUNT := SLAVE_COUNT; -- 1 more slave than connected for handling bad requests.
		variable slave_in_use : std_logic_vector(0 to SLAVE_COUNT);
	--		procedure calculate_slave(address : in std_logic_vector(ADDRESS_BITS - 1 downto 0)) is
	--		begin
	--			slave_idx := SLAVE_COUNT;
	--			for i in 0 to SLAVE_COUNT - 1 loop
	--				if (address and mask_array(i)) = address_array(i) then
	--					slave_idx := i;
	--				end if;
	--			end loop;
	--
	--		end procedure calculate_slave;

	begin
		if rst = '1' then
			for i in 0 to MASTER_COUNT - 1 loop
				aw_states(i) <= AW_READY;
			end loop;
			awready_s    <= (others => '1');
			slave_in_use := (others => '0');
		elsif rising_edge(clk) then
			for i in 0 to MASTER_COUNT - 1 loop -- Loop for every master
				case aw_states(i) is
					when AW_READY =>
						masters_in(i).awready <= '1';
						if masters_out(i).awvalid = '1' and awready_s(i) = '1' then -- check awready. just to prevent glitches
							slave_idx := calculate_slave(masters_out(i).awaddr, address_array, mask_array);
							if slave_in_use(slave_idx) /= '1' then
								write_locks_s(i).slave_idx   <= slave_idx;
								write_locks_s(i).locked      <= '1';
								slave_in_use(slave_idx)      := '1';
								awready_s(i)                 <= '0';
								-- output request to slave
								slaves_in(slave_idx).awaddr  <= masters_out(i).awaddr;
								slaves_in(slave_idx).awid    <= std_logic_vector(to_unsigned(i, WID_SLAVE_BITS - WID_MASTER_BITS)) & masters_out(i).awid;
								slaves_in(slave_idx).awburst <= masters_out(i).awburst;
								slaves_in(slave_idx).awcache <= masters_out(i).awcache;
								slaves_in(slave_idx).awlen   <= masters_out(i).awlen;
								slaves_in(slave_idx).awlock  <= masters_out(i).awlock;
								slaves_in(slave_idx).awprot  <= masters_out(i).awprot;
								slaves_in(slave_idx).awsize  <= masters_out(i).awsize;
								slaves_in(slave_idx).awvalid <= '1';
								-- Active state waits for awready to be asserted
								aw_states(i)                 <= AW_ACTIVE;
							end if;
						end if;
					when AW_ACTIVE =>
						if slaves_out(write_locks_s(i).slave_idx).awready = '1' then
							slaves_in(write_locks_s(i).slave_idx).awvalid <= '0';
							aw_states(i)                                  <= AW_BLOCK;
						end if;
					when AW_BLOCK =>
						if write_releases(i) = '1' then
							slave_in_use(write_locks_s(i).slave_idx) := '0';
							write_locks_s(i).locked                  <= '0';
							aw_states(i)                             <= AW_READY;
						end if;
				end case;

			end loop;
		end if;
	end process aw_router;

end architecture RTL;
